'''
 * this file is part of project
 * PIC4BBB
 * 
 * Program and control for dsPIC33EP / PIC24EP
 * with BeagleBone BLACK
 * M. 2017

 Free, open-source, fully featured demonstrative software.

THIS SOFTWARE IS PROVIDED,
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
PROVIDER BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

----------------------------------------------------------------------------------

This is script to program on fly the dsPIC33EP / PIC24EP on the PIC4BBB cape for
Beaglebone Black, with program in *.pbin file (see - pymcephex32.py).

The PRUs program is used dynamically - text.bin and data.bin.
It is different for the two types communication - via USART ot via I2C.
THIS SCRIPT IS SPECIALLY FOR I2C TYPE!

The programming process is reported with text and at very big PIC program,
can continue several decades of seconds.

Way of using - as command:

> python stdicsp.py name.pbin

At success, the PIC, i.e. the PIC4BBB cape is started as autonomous real-time
controller with communication to the main computer - Beaglebone Black.
'''

import pypruss as pru
import struct as st
import os, mmap, pickle, sys
from time import sleep


# HELPER functions for the Control Byte
# first_cell is 32 bits
# lower 24 bits are the address
# higher 1B is ctrlB - see PRU code

def toWrite(flg, first_cell):
    # flg = True / False to start or not writting
    if flg: return first_cell & 0xFEFFFFFF
    else: return first_cell | 0x1000000
def isWritting(first_cell):
    return False if first_cell & 0x1000000 else True

def isError(first_cell):
    return True if first_cell & 0x2000000 else False
def clrError(first_cell):
    return first_cell & 0xFDFFFFFF

def setType(tp, first_cell):
    if tp == 'pdata': return first_cell & 0xF3FFFFFF
    elif tp == 'iddata': return (first_cell & 0xF7FFFFFF) | 0x4000000
    elif tp == 'cdata': return (first_cell & 0xFBFFFFFF) | 0x8000000
    else: print('no such type'); raise Exception

def setCtrlB(tp, write_flg, first_cell):
    return clrError(setType(tp, toWrite(write_flg, first_cell)))


if (len(sys.argv) < 2) or (not sys.argv[1].endswith('.pbin')):
    print('enter PBIN file name')
    sys.exit(-1)

# load memory data for programming
with open(sys.argv[1], 'rb') as f:
    pdata, iddata, cdata = pickle.load(f)

os.system('echo mcape-universal > /sys/devices/bone_capemgr.9/slots') # Enable PRU
# config pins - in accordance with I2C circuit
os.system('config-pin P9_17 low') # PGEC out /I2C1_SCL
os.system('config-pin P9_18 in+') # PGED input with pull up /I2C1_SDA
os.system('config-pin P9_25 pruout') # PGEDout
os.system('config-pin P9_26 low') # MCLR out

PRUSS0_PRU0_DATARAM = 0
PRUSS0_PRU1_DATARAM = 1
PRUSS0_PRU0_IRAM = 2
PRUSS0_PRU1_IRAM = 3
PRUSS0_SHARED_DATARAM = 4

with open('data.bin', 'rb') as f:
    code_data = f.read()

out_code_data = st.unpack('<{}I'.format(len(code_data)/4), code_data)

pru.modprobe()
pru.init()
pru.open(0)
pru.pruintc_init()

# prepare first page for programming
p = len(pdata) # resting pdata pages for programming
pdata[0][0] = setCtrlB('pdata', False, pdata[0][0])
pru.pru_write_memory(PRUSS0_SHARED_DATARAM, 0, pdata[0])

pru.pru_write_memory(PRUSS0_PRU0_DATARAM, 0, out_code_data)
pru.exec_program(0, "./text.bin")  # the 0 is the number of PRU => use PRU0

# PRU start to program first page, check for complettion

isSucceed = False
with open("/dev/mem", "r+b") as f:
    ddr_mem = mmap.mmap(f.fileno(), 0x20000, offset=0x4a300000)

    def mem(cell):
        # just like PRU macros, but only for READING - each cell is 32 bit from the Shared RAM
        return st.unpack('<I', ddr_mem[0x10000 + cell*4:0x10004 + cell*4])[0]

    try:
        # programming pdata
        while True:
            print('program page {} / {} of pdata'.format(len(pdata)-p, len(pdata)))
            # start writting already set page
            pru.pru_write_memory(PRUSS0_SHARED_DATARAM, 0, [toWrite(True, mem(0))])

            # wait complettion
            while isWritting(mem(0)): pass

            # check for Error
            if isError(mem(0)):
                print(map(hex,[mem(i) for i in range(20)]))
                print('Writting Error'); raise Exception

            p -= 1 # prepare next page
            if p == 0: break # end

            pdata[len(pdata)-p][0] = setCtrlB('pdata', False, pdata[len(pdata)-p][0])
            pru.pru_write_memory(PRUSS0_SHARED_DATARAM, 0, pdata[len(pdata)-p])

        # programming iddata - one page
        print('program iddata')
        iddata[0] = setCtrlB('iddata', False, iddata[0])
        pru.pru_write_memory(PRUSS0_SHARED_DATARAM, 0, iddata)
        # start writting already set page
        pru.pru_write_memory(PRUSS0_SHARED_DATARAM, 0, [toWrite(True, mem(0))])
        # wait complettion
        while isWritting(mem(0)): pass
        # check for Error
        if isError(mem(0)): print('Writting Error'); raise Exception

        # programming cdata - one page
        print('program cdata')
        cdata[0] = setCtrlB('cdata', False, cdata[0])
        pru.pru_write_memory(PRUSS0_SHARED_DATARAM, 0, cdata)
        # start writting already set page
        pru.pru_write_memory(PRUSS0_SHARED_DATARAM, 0, [toWrite(True, mem(0))])
        # wait complettion
        while isWritting(mem(0)): pass
        # check for Error
        if isError(mem(0)): print('Writting Error'); raise Exception

        pru.wait_for_event(0)
        pru.clear_event(0, pru.PRU0_ARM_INTERRUPT)
        isSucceed = True
        print('successfull end')
    finally:
        # end
        # for i in range(20): print(i, hex(mem(i))); # can dump some memory for debug if necessary
        ddr_mem.close()
        pru.exit()

        # prepare pins for general PIC using with I2C communication
        os.system('config-pin P9_17 i2c') # PGEC to I2C1_SCL
        os.system('config-pin P9_18 i2c') # PGED to I2C1_SDA
        os.system('config-pin P9_25 low') # PGEDout - open collector - now can use PRU without to disturb the PIC
        os.system('config-pin P9_26 low') # MCLR out - reset PIC work - PIC is inactive if NOT succees reprogramming!!!

# restart of the PIC
if isSucceed:
    sleep(1.5)
    os.system('config-pin P9_26 high') # MCLR out - allow PIC work
    print('RUN PIC') # Here ALWAYS the reprogramming is fully successful and PIC4BBB cape start to
                     # execute it program and work as autonomous real-time controller for BBB
